-- kompilieren mit
--  ghc -o dpconsole --make DPConsole.hs
-- 
-- danach benutzen mit
-- ./dpconsole <FORMEL>
-- oder um die Formel aus einem File zu lesen
-- ./dpconsole -f <FILENAME>

module Main where

 import System.Environment
 import DavisPutnamschnell
 import Data.Char
 import Data.List
 import Data.Maybe

 helptext = "Usage: dpconsole [-fast] [-f|-k|-kf|-cnf|-cnff|-h] [Formel/Klauselmenge] [PATH]\n" ++
            "\n" ++
            "-----------------------------------------------------------------------------\n" ++
            "Aufruf-Varianten:\n" ++
            "dpconsole \"Formel\"              :\n"++
            "     prueft die Aussagenlogische Formel auf Erfuellbarkeit mit Davis-Putnam\n" ++
            "dpconsole -a \"Formel\"              :\n"++
            "     prueft die Aussagenlogische Formel auf Erfuellbarkeit mit Davis-Putnam\n" ++
            "     und berechnet ALLE Modelle\n" ++            
            "dpconsole -f dateipfad            :\n"++
            "     liest die Formel aus der Datei dateipfad und prueft dann die\n" ++
            "     aussagenlogische Formel auf Erfuellbarkeit mit Davis-Putnam\n" ++
            "dpconsole -af dateipfad            :\n"++
            "     liest die Formel aus der Datei dateipfad und prueft dann die\n" ++
            "     aussagenlogische Formel auf Erfuellbarkeit mit Davis-Putnam\n" ++
            "     und berechnet ALLE Modelle\n" ++            
            "dpconsole -k \"Klauselmenge\"     :\n"++
            "     prueft die Klauselmenge mit DP\n"++
            "dpconsole -kf dateipfad           :\n"++
            "     liest die Klauselmenge aus der Datei dateipfad und wendet DP an \n" ++
            "dpconsole -k \"Klauselmenge\"     :\n"++
            "     prueft die Klauselmenge mit DP\n"++
            "dpconsole -cnf \"Formel\"         :\n"++
            "     berechnet die Klauselnormalform der Formel\n"++
            "dpconsole -cnff dateipfad         :\n"++
            "     berechnet die Klauselnormalform der Formel in der Datei dateipfad\n"++
            "Die Option -fast verwendet die schnelle CNF-Berechnung\n"++
            "dpconsole -h oder dpconsole --help:\n"++
            "     zeigt diese Hilfe an\n" ++
            "-----------------------------------------------------------------------------\n" ++
            "Beispiele:\n"++
            "dpconsole \"A \\/ B /\\ (-A <=> B)\"\n"++
            "dpconsole -cnf \"A \\/ B /\\ (-A <=> B)\"\n"++
            "dpconsole -k  \"[[1,-2],[1,3],[-2,3]]\""

 

 data Mode = PF | DP | CK | H | PFA

 main   = do
    args <- getArgs
    let (fast,args') = checkfast args
    (inp,mode) <- checkargs args'
    case inp of 
     Nothing -> case mode of 
                  H ->  putStrLn helptext
                  other -> return ()
     Just a -> case mode of 
                PFA -> 
                  do
                   let format = (filter (not.isSpace) a)
                   putStrLn "Die Formel ist: "
                   putStrLn format
                   putStrLn ""
                   putStrLn (unlines $ map (\x -> "Modell:" ++ (concat . intersperse ",") x) $ (if fast then solveAllFast else solveAll) format )                    
                PF -> 
                  do
                   let format = (filter (not.isSpace) a)
                   putStrLn "Die Formel ist: "
                   putStrLn format
                   putStrLn ""
                   putStrLn $ (if fast then dpfcnf else dp) format 
                DP ->  
                   do
                    if fast then 
                     putStrLn "ignoring option -fast"
                            else return ()
                    let dpa = map sort $ parseKlauselMenge (filter (not.isSpace) a)
                    putStrLn (showRes $ davisPutnamSat dpa [])
                CK ->  
                  do
                   let format = (filter (not.isSpace) a)
                   putStrLn "Die Formel ist: "
                   putStrLn format
                   putStrLn ""
                   let (kmenge,subs) = ((if fast then klauselmengef else klauselmenge) format) :: ([[Int]],[(String,Int)])
                   putStrLn $ "Die errechnete Klauselmenge in CNF ist: \n" ++ b60 (clear (show (map (map (subst (invsubs subs))) kmenge)))


--   (klauselmengef format )  ::([[Int]],[(String,Int)])


 klauselmengef inp =  let (subs,e) = numPropSubs . fastcnfInt . parseProp $ inp
                      in 
                         let
                          inpcnf =  cnf e
                          res =  dpcnf inpcnf 
                         in (res, subs)

 klauselmenge inp =  let (subs,e) = numPropSubs . parseProp $ inp
                     in 
                         let
                          inpcnf =  cnf e
                          res =  dpcnf inpcnf 
                         in (res, subs)

 invsubs = map (\a ->(snd a,fst a))

 subst  list x
  | x < 0 = "-" ++ (fromJust $ lookup (abs x) list)
  | x > 0 = fromJust $ lookup x list

 checkfast ("-fast":args) = (True,args)
 checkfast xs             = (False,xs)
   
 checkargs []          = do
                          putStrLn "Missing Input!\n"
                          return (Nothing,H)
                          
 checkargs ("-f":args) = do
                           content <- readFile (head args)
                           return ((Just content),PF)

 checkargs ("-af":args) = do
                           content <- readFile (head args)
                           return ((Just content),PFA)
                           
 checkargs ("-k":args) = do
                           return ((Just (head args)), DP)

 checkargs ("-kf":args) = do
                           putStrLn ("Reading file: " ++ head args)
                           content <- readFile (head args)
                           return ((Just content),DP)

 checkargs ("-cnf":args) = do
                            return ((Just (head args)), CK)

 checkargs ("-cnff":args) = do
                             putStrLn ("Reading file: " ++ head args)
                             content <- readFile (head args)
                             return ((Just content),CK)

 checkargs ("-h":args) = do
                             return (Nothing,H)

 checkargs ("--help":args) = do
                               return (Nothing,H)


 checkargs ("-a":(a:args)) = do
                             return (Just a, PFA)

 checkargs (a:args)    = return ((Just a),PF)

 --
 -- Klauselparser
 --
 
 showRes list =  let str = concat (intersperse ", " (map show list)) in "[" ++ str ++ "]"

 parseKlauselMenge [] = []
 parseKlauselMenge ('[':xs) = let (inKl,res) = findInKlammer 0 xs in 
                                if filter (not.isSpace) res /= [] then error "Fehler: ungueltiges Format" else
                                  parse inKl

 findInKlammer :: Int -> String -> (String,String)
 findInKlammer 0 (']': xs) = ([],xs)
 findInKlammer i (']': xs) = let (ink,r) = findInKlammer (i-1) xs in (']':ink,r)
 findInKlammer i ('[': xs) = let (ink,r) = findInKlammer (i+1) xs in ('[':ink,r)
 findInKlammer i (x:xs)    = let (ink,res)= findInKlammer i xs in (x:ink,res)
 findInKlammer _ _ = error "Fehler: ungueltiges Format"
  

 parse []        = []
 parse ('[':xs) = let (klausel,rest) = parseKlammer xs
                  in parseKlausel klausel : parse rest
 parse (x:xs)
  | isSpace x  = parse xs
  | x == ','   = parse xs
  | otherwise  = error $ show (x:xs)
 
 parseKlausel [] = []
 parseKlausel (x:xs)
   | isSpace x = parseKlausel xs
   | x == ','  = parseKlausel xs
   | otherwise = let (num,res) = parseNum (x:xs)
                 in num:(parseKlausel res)

 parseNum ('-':xs) = let (num,res) = parseNum xs
                    in ((- num),res)
 parseNum xs = let (num,res) = break (not.isDigit) xs
               in if null num then error "Fehler: Nur Zahlen sind als Literale erlaubt!" else ((read num)::Int,res)

 parseKlammer []       = error "Fehler: fehlende ']'"  
 parseKlammer (']':xs) = ([],xs)
 parseKlammer (x:xs)   = let (inK,res) = parseKlammer xs
                         in  (x:inK,res)

 clear [] = []
 clear ('\"':xs) = clear xs
 clear (x:xs) = x:(clear xs)

 b60 = unlines . break60
 break60 [] = []
 break60 xs = (take 60 xs) : (break60 (drop 60 xs))

 
 solveAllFast inp     =

           let  (l,_) = numPropSubs $ parseProp $ inp
                n = map fst l
                (subs,e) = numPropSubs $ fastcnfInt $  parseProp $ inp
                inpcnf =  (cnf e)::(Pexpr Int)
                variables = fnub $ sort (map snd subs)
                allm = iterateD  (dpcnf $ inpcnf) variables
                res = solver . dpcnf $ inpcnf                 

           in  -- trace (show inp) $ trace (show e) $
                if   inpcnf == (Pand [])     
                  then  []
                 else
                if null res  
                then []
                else  (map (\r -> (backSubs1 n r subs))  $ fnub $ map sort $ allm)
  where
   iterateD cnf vars =
           let
                res = (solver cnf)
                res' = res ++ (vars \\ (map abs res))
           in if null res then []
                        else 
                          (res':(iterateD (cnf ++ [map (*(-1)) res']) vars)) 
   fnub (x:y:xs) 
    | x == y = fnub (x:xs)
    | otherwise = x: (fnub (y:xs))
   fnub xs = xs
   solver = davisPutnam
   backSubs1 names [] subs = []
   backSubs1 names (x:xs) subs = 
    case doBackSubs x subs  of
     Just i -> i : backSubs1 names xs subs 
     Nothing -> backSubs1 names xs subs
    where 
      doBackSubs x [] = Nothing
      doBackSubs x ((a,n):arest) = 
        if n == x && a `elem` names then Just (a)
        else if -n == x && a `elem` names
            then Just ("-" ++ a)
            else doBackSubs x  arest      
                         
                         
 solveAll inp     =

           let  (l,_) = numPropSubs $ parseProp $ inp
                n = map fst l
                -- (subs,e) = numPropSubs $ fastcnfInt $  parseProp $ inp
                (subs,e) = numPropSubs . parseProp $ inp
                inpcnf =  (cnf e)::(Pexpr Int)
                variables = fnub $ sort (map snd subs)
                allm = iterateD  (dpcnf $ inpcnf) variables
                res = solver . dpcnf $ inpcnf                 

           in  
                if   inpcnf == (Pand [])     
                  then  []
                 else
                if null res  
                then []
                else  (map (\r -> (backSubs1 n r subs))  $ fnub $ map sort $ allm)
  where
   iterateD cnf vars =
           let
                res = (solver cnf)
                res' = res ++ (vars \\ (map abs res))
           in if null res then []
                        else 
                          (res':(iterateD (cnf ++ [map (*(-1)) res']) vars)) 
   fnub (x:y:xs) 
    | x == y = fnub (x:xs)
    | otherwise = x: (fnub (y:xs))
   fnub xs = xs
   solver = davisPutnam
   backSubs1 names [] subs = []
   backSubs1 names (x:xs) subs = 
    case doBackSubs x subs  of
     Just i -> i : backSubs1 names xs subs 
     Nothing -> backSubs1 names xs subs
    where 
      doBackSubs x [] = Nothing
      doBackSubs x ((a,n):arest) = 
        if n == x && a `elem` names then Just (a)
        else if -n == x && a `elem` names
            then Just ("-" ++ a)
            else doBackSubs x  arest      
                          